module Text.Safe.Strings.TH ( q, declareSafeString ) where

import Text.Safe.Strings.Core

import Language.Haskell.TH
import Language.Haskell.TH.Syntax
import qualified Text.ParserCombinators.ReadP as P
import Control.Monad

-- Convert template sugar into calls to frag, text, cat, etc.
-- This function is exported by the SafeStrings module.

q :: String -> ExpQ
q spec =
    [| cat $(parts) |]
  where
    parts = case xparse spec of
        []   -> error ("bad template: " ++ show spec)
        ps:_ -> foldr gen [| [] |] ps
    gen p ps' = (\p' -> [| $p' : $ps' |]) $ case p of
        SFrag s  -> [| frag $(litE (stringL s))         |]
        SIFrag s -> [| $(varE (mkName s))               |]
        SIShow s -> [| text (show $(varE (mkName s)))   |]
        SITxt s  -> [| text $(varE (mkName s))          |]
        SIRTxt s -> [| text (render $(varE (mkName s))) |]

-- AST for template-specification parts

data SpecPart
    = SFrag String  -- ^ language fragment
    | SIFrag String -- ^ insert fragment by variable reference
    | SIShow String -- ^ insert rendered variable via show
    | SITxt String  -- ^ insert literal text variable
    | SIRTxt String -- ^ insert rendered safe string var as text
  deriving Show

-- Parse a template specification

xparse :: String -> [[SpecPart]]
xparse spec = do
    (result, "") <- P.readP_to_S templateP spec
    return result
 where
    templateP = do
        P.many ((liftM SFrag (P.munch1 (/= '#'))) P.<++
                interpolationP P.<++
                liftM SFrag (P.string "#"))

    interpolationP = do
        P.string "#{"
        spec <- P.manyTill P.get (P.char '}')
        return $ case spec of
          'r':' ':var -> SIRTxt (strip var)
          's':' ':var -> SIShow (strip var)
          '=':var     -> SIFrag (strip var)
          var         -> SITxt  (strip var)

strip = frontAndBack (dropWhile (== ' '))
frontAndBack f = reverse . f . reverse . f

{-
 - psha: Don't know if it's correct from Moertel's point of view
 - but it's working :)
 -}
declareSafeString :: String -> String -> ExpQ -> Q [Dec]
declareSafeString lang name expr = let
	ss = mkName "SafeString"
	lss = mkName "LocalSafeString" 
        ssType = AppT (ConT ss) (ConT lss)
	in [d|
	newtype LocalSafeString 
	    = LocalSafeString { unLocalSafeString :: String }
	    deriving Show

	instance Language LocalSafeString where
	    litfrag  = LocalSafeString
	    littext  = LocalSafeString . $(expr)
	    natrep   = unLocalSafeString
	    language = const $(stringE lang)

	--type Url = SafeString LocalSafeString
	|] >>= (\x -> return ( TySynD 
                               (mkName name) 
                               [] 
                               ssType
                             : ValD
                                    (VarP (mkName lang))
                                    ( NormalB
                                      ( SigE
                                        (VarE (mkName "frag"))
                                        ( AppT
                                          (AppT ArrowT (ConT (mkName "String")))
                                          ssType
                                        )
                                      )
                                    )
                                   []
                             : ValD
                                    (VarP (mkName ("render" ++ name)))
                                    ( NormalB
                                      ( SigE
                                        (VarE (mkName "render"))
                                        ( AppT
                                          (AppT ArrowT ssType)
                                          (ConT (mkName "String"))
                                        )
                                      )
                                    )
                                   []
                             : x)
               )
	{-
	 let t = [d|type D a = Int; type D' = D Int |]
	 (runQ t) >>= (putStrLn.show)
	 -}
